from enum import Enum, unique


@unique
class Rebalance(str, Enum):
    JOIN = "join"
    LEAVE = "leave"
    DEAD = "dead"

    def message(self, **kwargs):
        return dict(reason=self.value, **kwargs)


@unique
class ExceptionPolicy(str, Enum):
    """
    An enum specifying how to handle exceptions raised in the user-provided processor
    coroutine.

    * ``HALT``: Raise the exception, halting execution of the affected partition.
    * ``QUARANTINE``: Mark the affected partition as poisoned, and continue with others.
    * ``IGNORE``: Suppress the exception and continue processing regardless.

    Examples
    --------
    >>> @app.processor(orders, exception_policy=ExceptionPolicy.HALT)
    ... async def printer(events):
    ...     async for order in events.records():
    ...         print(order.amount)
    """
    HALT = "halt"
    QUARANTINE = "quarantine"
    IGNORE = "ignore"



from functools import lru_cache
from hashlib import md5

try:
    from xxhash import xxh64
except ImportError:
    xxh64 = None


@lru_cache
def md5hash(x):
    return int(md5(repr(x).encode("utf-8")).hexdigest(), 16)


@lru_cache
def fasthash(x):
    return xxh64(repr(x)).intdigest()


if xxh64:
    default = fasthash
else:
    default = md5hash



import datetime
import logging
import logging.config
import os
from io import StringIO
from typing import Dict

import colorama
import structlog

from runnel.context import executor_id, partition_id, stream, worker_id


def init_logging(level="debug", format="console"):
    level = level.upper()

    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "standard": {
                "format": "%(message)s"
            }
        },
        "handlers": {
            "default": {
                "level": level,
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "": {
                "handlers": ["default"],
                "level": level,
            },
            "runnel": {
                "handlers": ["default"],
                "level": level,
                "propagate": False,
            },
        },
    })

    structlog.configure(
        logger_factory=logging.getLogger,
        processors=[
            _add_meta,
            structlog.stdlib.filter_by_level,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer() if format == "json" else _renderer,
        ],
    )


def _add_meta(logger: str, method_name: str, event_dict: Dict) -> Dict:
    event_dict = {
        "event": event_dict.pop("event"),
        "exc_info": event_dict.pop("exc_info", None),
        "extra": event_dict,
    }

    if not event_dict["extra"]:
        event_dict.pop("extra")

    event_dict["timestamp"] = datetime.datetime.utcnow().isoformat() + "Z"
    event_dict["level"] = method_name.upper()
    event_dict["pid"] = os.getpid()
    event_dict["worker_id"] = worker_id.get()
    event_dict["executor_id"] = executor_id.get()
    event_dict["stream"] = stream.get()
    event_dict["partition_id"] = partition_id.get()
    return event_dict


reset = colorama.Style.RESET_ALL
bright = colorama.Style.BRIGHT
dim = colorama.Style.DIM

red = colorama.Fore.RED
yellow = colorama.Fore.YELLOW
blue = colorama.Fore.BLUE
green = colorama.Fore.GREEN

colours = {
    "CRITICAL": red + bright,
    "EXCEPTION": red + bright,
    "ERROR": red + bright,
    "WARNING": yellow + bright,
    "INFO": green + bright,
    "DEBUG": blue + bright,
}


def _renderer(logger: str, method_name: str, event_dict: Dict) -> str:
    s = StringIO()

    ts = event_dict.pop("timestamp", None)
    if ts:
        s.write(f"{reset}{dim}{str(ts)}{reset} ")

    level = event_dict.pop("level", None)
    if level:
        s.write(f"{colours[level]}{level:>9}{reset} ")

    event = event_dict.pop("event")
    s.write(f"{bright}{event:<17}{reset} ")

    worker_id = event_dict.pop("worker_id", None)
    if worker_id:
        s.write(f"{blue}worker_id{reset}={worker_id} ")

    executor_id = event_dict.pop("executor_id", None)
    if executor_id:
        s.write(f"{blue}executor_id{reset}={executor_id} ")

    stream = event_dict.pop("stream", None)
    partition_id = event_dict.pop("partition_id", None)
    suffix = f"-{partition_id}" if partition_id is not None else ""
    if stream:
        s.write(f"{blue}stream{reset}={stream+suffix} ")

    for key, value in event_dict.pop("extra", {}).items():
        s.write(f"{blue}{key}{reset}={value}{reset} ")

    stack = event_dict.pop("stack", None)
    exc = event_dict.pop("exception", None)

    if stack is not None:
        s.write("\n" + stack)
        if exc is not None:
            s.write("\n\n" + "=" * 88 + "\n")
    if exc is not None:
        s.write("\n" + exc)

    return s.getvalue()



import json
import random
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, Set

import anyio
import structlog

from runnel.constants import Rebalance
from runnel.context import executor_id, stream
from runnel.exceptions import (
    AcquirePartitionLockFailed,
    ExtendPartitionLockFailed,
    NoPartitionsAssigned,
    RebalanceFailed,
)
from runnel.runner import Runner
from runnel.types import Partition

if TYPE_CHECKING:
    from runnel.processor import Processor

logger = structlog.get_logger(__name__)


@dataclass
class Executor:
    """
    This object is responsible for executing the user-provided processor function. It
    expects to join a distributed set of other executors and coordinates ownership
    over a subset of Redis stream partitions.
    """
    id: str
    processor: "Processor"
    partitions: Dict[int, Partition] = field(default_factory=dict)
    cond: anyio.abc.Condition = field(default_factory=anyio.create_condition)
    assigned: Set[Partition] = None
    owned: Set[Partition] = field(default_factory=set)
    poisoned: Set[Partition] = field(default_factory=set)

    def __post_init__(self):
        for i in range(self.stream.partition_count):
            self.partitions[i] = Partition(stream=self.stream, number=i)

    def __hash__(self):
        return hash((self.processor.name, self.id))

    @property
    def app(self):
        return self.processor.app

    @property
    def stream(self):
        return self.processor.stream

    @property
    def group(self):
        return self.processor.group

    @property
    def consumer(self):
        return self.processor.consumer

    @property
    def lock_expiry(self):
        return self.processor.lock_expiry

    def redis_lock_key(self, i):
        return self.processor.lock_key(i)

    @property
    def safe_partitions(self):
        return self.owned - self.poisoned - self.need_release

    @property
    def need_release(self):
        return self.owned - (self.assigned or set())

    @property
    def need_acquire(self):
        return (self.assigned or set()) - self.owned

    @property
    def rebalanced(self):
        return self.owned == self.assigned

    async def set_assigned(self, partitions: Set):
        async with self.cond:
            self.assigned = partitions
            await self.cond.notify_all()

    async def set_owned(self, partitions: Set):
        async with self.cond:
            self.owned = partitions
            await self.cond.notify_all()

    async def wait_for(self, fn=None, attr=None, value=None):
        if not fn:
            fn = lambda: getattr(self, attr) == value

        async with self.cond:
            while not fn():
                await self.cond.wait()

    @asynccontextmanager
    async def run_locks(self, partitions):
        for p in partitions:
            await p.lock.acquire()
        try:
            yield
        finally:
            for p in partitions:
                await p.lock.release()

    async def start(self):
        try:
            executor_id.set(self.id)
            stream.set(self.stream.name)
            await self._beat()
            await self.processor.maybe_join(self.id)
            await anyio.sleep(self.processor.join_delay)
            logger.info("executor-start")

            async with anyio.create_task_group() as tg:
                await tg.spawn(self.heartbeat)
                await tg.spawn(self.maintenance)
                await tg.spawn(self.control)
                await tg.spawn(Runner(self).run)
        finally:
            # Ensure that our partition locks are released and our departure is
            # broadcast to other workers.
            await self.release(set(self.owned), suppress_cancelled=True)
            await self.processor.maybe_leave(self.id)
            logger.critical("executor-exit")

    async def heartbeat(self):
        while True:
            await self._beat()
            await self.extend()
            await anyio.sleep(self.processor.lock_expiry / 4)

    async def _beat(self):
        logger.debug("beating")
        px = self.processor.lock_expiry * 1000
        await self.app.redis.set(self.processor.heartbeat_key(self.id), "beat", px=px)

    async def maintenance(self):
        while True:
            logger.info("maintenance")
            await self.processor.reap()
            jitter = self.processor.lock_expiry / 2 * random.random() + 1
            await anyio.sleep(jitter)

    async def control(self):
        send_stream, receive_stream = anyio.create_memory_object_stream()
        async with anyio.create_task_group() as tg:
            await tg.spawn(self._message_listener, send_stream)
            await tg.spawn(self._message_consumer, receive_stream)

    async def _message_listener(self, send_stream):
        async for msg in self.processor.read_control_messages():
            if msg and msg[b"reason"].decode("utf-8") in set(Rebalance):
                logger.debug("control-received", msg=msg)
                await send_stream.send(msg)

    async def _message_consumer(self, receive_stream):
        attempts = 0

        while True:
            # Wait for a control message from the queue, but proceed anyway
            # with a rebalance check after a long sleep.
            async with anyio.move_on_after(self.processor.assignment_sleep):
                await receive_stream.receive()

            try:
                completed = await self.maybe_rebalance()
            except NoPartitionsAssigned:
                # A configuration error -- there are more group members than stream
                # partitions. Wait until one becomes available.
                logger.warning("no-partitions-assigned")
                attempts = 0
            except AcquirePartitionLockFailed:
                # Our declared partition assignment is out of date (e.g. a new executor
                # has been added since we read it) or the current owner has not released
                # it yet, so we cannot acquire the locks for our declared partitions.
                attempts += 1
                if attempts >= self.processor.assignment_attempts:
                    # Either our partition assignment is incorrect or another worker is
                    # failing to respond to control messages to release its partition
                    # locks. Give up, killing this process.
                    logger.critical("rebalance-failed")
                    raise RebalanceFailed
                else:
                    # It's likely that the current owner is still processing its backlog
                    # of events and will release it soon.
                    logger.warning("could-not-acquire")
            else:
                if completed:
                    logger.info("rebalance-completed")
                    attempts = 0
                else:
                    logger.debug("rebalance-check-passed")

    async def maybe_rebalance(self):
        await self.processor.reap()
        result = await self.stream.app.redis.get(self.processor.members_key)
        members = json.loads(result)

        # The only way we're not a declared member is if we were deemed dead because our
        # heartbeat expired. This should never happen.
        assert self.id in members

        assigned = set(self.partitions[i] for i in members[self.id]["partitions"])
        await self.set_assigned(assigned)

        acquiring = self.need_acquire
        releasing = self.need_release

        if acquiring or releasing:
            logger.warning(
                "rebalancing",
                assigned=[p.number for p in assigned],
                owned=[p.number for p in self.owned]
            )

        if releasing:
            logger.info("waiting-to-release", partitions=[p.number for p in releasing])
            async with self.run_locks(releasing):
                await self.release(releasing)

        if acquiring:
            logger.info("waiting-to-acquire", partitions=[p.number for p in acquiring])
            async with self.run_locks(acquiring):
                await self.acquire(acquiring)

        if not assigned:
            assert len(members) >= self.stream.partition_count
            raise NoPartitionsAssigned

        if acquiring or releasing:
            return True

    async def acquire(self, partitions):
        client = self.app.redis
        keys = [self.redis_lock_key(p.number) for p in partitions]
        px = self.lock_expiry * 1000

        async with await client.pipeline(transaction=True) as pipe:
            for key in keys:
                await pipe.set(key, self.id, nx=True, px=px)
            acquired = await pipe.execute()
            failed = False

        new = set()
        for p, success in zip(partitions, acquired):
            if success:
                p.reset()
                new.add(p)
            else:
                failed = True
        await self.set_owned(self.owned | new)

        if failed:
            logger.debug("failed-to-acquire", result=acquired)
            raise AcquirePartitionLockFailed
        logger.debug("acquired-partitions", partitions=[p.number for p in partitions])

    async def release(self, partitions, suppress_cancelled=False):
        client = self.app.redis
        fn = self.app.scripts["lock_release"]
        keys = [self.redis_lock_key(p.number) for p in partitions]

        async with await client.pipeline(transaction=True) as pipe:
            for key in keys:
                await fn.execute(client=pipe, keys=[key], args=[self.id])

            # Shield the execution from cancellation to ensure the locks are released.
            async with anyio.fail_after(4, shield=True):
                await pipe.execute()

        try:
            await self.set_owned(self.owned - partitions)
        except anyio.get_cancelled_exc_class():
            if not suppress_cancelled:
                raise
        finally:
            logger.debug("released-partitions", partitions=[p.number for p in partitions])

    async def extend(self):
        if self.owned:
            client = self.app.redis
            fn = self.app.scripts["lock_extend"]
            keys = [self.redis_lock_key(p.number) for p in self.owned]
            px = self.lock_expiry * 1000

            async with await client.pipeline(transaction=True) as pipe:
                for key in keys:
                    await fn.execute(client=pipe, keys=[key], args=[self.id, px])
                extended = await pipe.execute()

            if not all(extended):
                raise ExtendPartitionLockFailed
            logger.debug("extended-partitions", partitions=[p.number for p in self.owned])



import structlog

from runnel.interfaces import Middleware
from runnel.types import Event

logger = structlog.get_logger(__name__)


class Deserialize(Middleware):
    async def handler(self, parent, events, **kwargs):
        async for x in parent:
            assert isinstance(x, (Event, list))

            if isinstance(x, Event):
                yield events.stream.deserialize(x.data)
            else:
                yield [events.stream.deserialize(event.data) for event in x]



# yapf: disable
from datetime import datetime
from typing import Iterable

import structlog

from runnel.interfaces import Middleware
from runnel.record import Record
from runnel.types import Event

logger = structlog.get_logger(__name__)


# An example record to demonstrate forwarding of failed events to a dead letter queue.
# Stores the raw bytes since deserialization may have been responsible for the failure.
class Dead(Record, primitive=True):
    failed_at: str
    partition: str
    xid: str
    data: bytes

    @classmethod
    def from_event(cls, event: Event):
        return cls(
            failed_at=datetime.utcnow().isoformat(),
            partition=event.partition.key,
            xid=event.xid,
            data=event.data[b"data"],
        )


class DeadLetterMiddleware(Middleware):
    def __init__(self, source):
        self.source = source
        name = f"dead.{source.name}"

        if source.record._primitive:
            # Re-use the source record type as the dead record type so we don't
            # have to perform any serialization/deserialization.
            self.sink = source.clone(name=name, partition_count=1)
        else:
            # Use the example Dead type.
            self.sink = source.app.stream(
                name,
                record=Dead,
                partition_by="partition",
                partition_count=1,
            )

    async def handler(self, parent, **kwargs):
        async for x in parent:
            assert isinstance(x, (Event, list))
            try:
                yield x
            except GeneratorExit:
                try:
                    logger.warning("dead-letter-handling", x=x)
                    if isinstance(x, Event):
                        await self._send(x)
                    else:
                        await self._send(*x)
                except:
                    logger.exception("dead-letter-failed")
                raise

    async def _send(self, *events: Iterable[Event]):
        if self.source.record._primitive:
            # Convert the event back into the source record type.
            convert = lambda e: self.sink.record(**{k.decode("utf-8"): v for k, v in e.data.items()})
        else:
            convert = lambda e: Dead.from_event(e)

        # Forward to the sink stream.
        await self.sink.send(*[convert(e) for e in events])



from typing import Callable, Coroutine, Dict, Set

import anyio
import structlog
from aredis import StrictRedis

from runnel import context
from runnel.exceptions import Misconfigured
from runnel.logging import init_logging
from runnel.processor import Processor
from runnel.settings import Settings
from runnel.stream import Stream
from runnel.utils import seconds_until
from runnel.worker import Worker

logger = structlog.get_logger(__name__)


class App:
    """
    This is the main abstraction provided by Runnel. Use the :meth:`~.stream` and
    :meth:`~.processor` methods to define partitioned event streams and processors
    respectively. Apps will be run by workers via the CLI.

    Parameters
    ----------
    name : str
        An app identifier. Will be used as a component of the Redis keys for the
        under-the-hood data structures.
    kwargs : Dict
        Any of the settings found in :mod:`runnel.settings` to override any options
        provided by environment variables.

    Examples
    --------
    >>> from runnel import App, Record
    ...
    >>> app = App(
    ...     name="example",
    ...     log_level="info",
    ...     redis_url="127.0.0.1:6379",
    ... )

    Specify your event types using the Record class:

    >>> class Order(Record):
    ...     order_id: int
    ...     created_at: datetime
    ...     amount: int
    ...     item_ids: List[int]

    Streams are configured to be partitioned by a chosen key.

    >>> orders = app.stream("orders", record=Order, partition_by="order_id")

    Processor functions iterate over the event stream and can rely on receiving events
    in the order they were created per key.

    >>> @app.processor(orders)
    ... async def printer(events):
    ...     async for order in events.records():
    ...         print(order.amount)

    Under the hood, Runnel will take care of partitioning the stream to enable scalable
    distributed processing. We also take care of concurrently running the async processor
    functions -- one for every partition in your stream. This processing may be
    distributed across many workers on different machines and Runnel will coordinate
    ownership of partitions dynamically as workers join or leave.
    """
    def __init__(self, name: str, **kwargs):
        self.name: str = name
        self.settings: Settings = Settings(**kwargs)
        self.redis = StrictRedis.from_url(self.settings.redis_url)
        self.workers: Set[Worker] = set()
        self.tasks: Set[Coroutine] = set()
        self.processors: Dict[str, Coroutine] = {}
        self.scripts: Dict[str, Callable] = {}

        init_logging(
            level=self.settings.log_level,
            format=self.settings.log_format,
        )

    @property
    def is_leader(self):
        return context.is_leader.get()

    def stream(
        self,
        name,
        record,
        partition_by,
        serializer=None,
        partition_count=None,
        partition_size=None,
        hasher=None,
    ):
        """
        A set of partitioned Redis streams, containing events as structured Record types.

        Kwargs, if provided, will override the default settings configured on the App
        instance or via environment variables (see :mod:`runnel.settings`) for this
        stream.

        Parameters
        ----------
        name : str
            An stream identifier. Will be used as a component of the Redis keys for the
            under-the-hood data structures.
        record: Type[Record]
            A class that inherits from Record, which specifies the structure of the event
            data this stream expects. See :class:`runnel.Record`.
        partition_by : Union[str, Callable[Record, Any]]
            A str representing an attribute of the Record type (or a callable to compute
            a value) which should be used to partition events. For example, if your events
            concern user activity and you want to process events in-order per user, you
            might choose the "user_id" attribute to partition by.
        serializer : Serializer
            An object implementing the :class:`runnel.interfaces.Serializer` interface
            which controls how records are stored in the Redis streams.
        partition_count : int
            How many partitions to create.
        partition_size : int
            The max length of each partition. (Implemented approximately via Redis' MAXLEN
            option to XACK.) Represents the size of the buffer in case processors are
            offline or cannot keep up with the event rate.
        hasher: Callable[Any, int]
            A function used to hash the partition key to decide to which partition to
            send a record.

        Examples
        --------
        >>> from runnel import App, Record, JSONSerializer
        ...
        >>> app = App(name="example")
        ...
        >>> class Order(Record):
        ...     order_id: int
        ...     amount: int

        Streams are configured to be partitioned by a chosen key.

        >>> orders = app.stream(
        ...     name="orders",
        ...     record=Order,
        ...     partition_by="order_id",
        ...     partition_count=16,
        ...     serializer=JSONSerializer(),
        ... )
        """
        if serializer is None and not record._primitive:
            serializer = self.settings.default_serializer

        return Stream(
            app=self,
            name=name,
            record=record,
            partition_by=partition_by,
            serializer=serializer,
            partition_count=partition_count or self.settings.default_partition_count,
            partition_size=partition_size or self.settings.default_partition_size,
            hasher=hasher or self.settings.default_hasher,
        )

    def processor(
        self,
        stream,
        *,
        name=None,
        exception_policy=None,
        middleware=None,
        lock_expiry=None,
        read_timeout=None,
        prefetch_count=None,
        assignment_attempts=None,
        assignment_sleep=None,
        grace_period=None,
        pool_size=None,
        join_delay=None
    ):
        """
        A wrapper around an async Python function which iterates over a continuous event
        stream.

        Kwargs, if provided, will override the default settings configured on the App
        instance or via environment variables (see :mod:`runnel.settings`) for this
        processor.

        Notes
        -----
        Events are acknowledged at the end of every processing loop. This means that if
        your processor crashed before completion, that section of work will be repeated
        when the processor is restarted. Therefore Runnel provides 'at least once'
        semantics.

        Parameters
        ----------
        stream : runnel.App.stream
            The stream this processor will iterate over.
        name : str
            Used in the Redis keys relating to this processor. Must be unique together
            with the App and Stream. Default: your function's ``__name__``.
        exception_policy : ExceptionPolicy
            How to handle exceptions raised in the user-provided processor coroutine.

            * ``HALT``: Raise the exception, halting execution of the affected partition.
            * ``QUARANTINE``: Mark the affected partition as poisoned, and continue with others.
            * ``IGNORE``: Suppress the exception and continue processing regardless.

            Default: ``HALT``.
        middleware : List[Middleware]
            A list of Middleware objects for managaing the data pipeline. Can be used to
            implement custom exception handling (e.g. dead letter queues).
        lock_expiry : int (seconds)
            The duration of the lock on stream partitions owned by executors of this
            processor. This controls the worst case lag a partition's events may
            experience since other executors will have to wait acquire the lock in case
            the owner has died.
        read_timeout : int (milliseconds)
            How long to stay blocked reading from Redis via XREADGROUP. Nothing depends
            on this.
        prefetch_count : int
            The maximum number of events to read from Redis per partition owned by an
            executor. (If a single executor owns all 16 partitions in a stream and
            prefetch_count is 10, then 160 events may be read at once.) Purely an
            optimisation.
        assignment_attempts : int
            How many times to try to complete a rebalance operation (i.e. acquire our
            declared partitions) before giving up.
        assignment_sleep : float (seconds)
            How long to wait between attempts to complete a rebalance operation.
        grace_period : float (seconds)
            How long to wait for execution to complete gracefully before cancelling it.
        pool_size : int
            How many concurrent connections to make to Redis to read events.
        join_delay : int (seconds)
            How long to wait after joining before attempting to acquire partitions.
            Intended to mitigate a thundering herd problem of multiple workers joining
            simultaneously and needing to rebalance multiple times.

        Examples
        --------
        >>> from runnel import App, Record
        ...
        >>> app = App(name="example")
        ...
        >>> class Order(Record):
        ...     order_id: int
        ...     amount: int
        ...
        >>> orders = app.stream(name="orders", record=Order, partition_by="order_id")
        ...
        >>> @app.processor(orders)
        ... async def printer(events):
        ...     async for order in events.records():
        ...         print(order.amount)
        """
        if not isinstance(stream, Stream):
            raise Misconfigured("You must pass a stream to the app.processor decorator")

        kwargs = {
            "exception_policy": exception_policy or self.settings.default_exception_policy,
            "middleware": middleware or [],
            "lock_expiry": lock_expiry or self.settings.default_lock_expiry,
            "read_timeout": read_timeout or self.settings.default_read_timeout,
            "prefetch_count": prefetch_count or self.settings.default_prefetch_count,
            "assignment_attempts": assignment_attempts or self.settings.default_assignment_attempts,
            "assignment_sleep": assignment_sleep or self.settings.default_assignment_sleep,
            "grace_period": grace_period or self.settings.default_grace_period,
            "pool_size": pool_size or self.settings.default_pool_size,
            "join_delay": join_delay or self.settings.default_join_delay,
        }

        def decorator(f):
            nonlocal name
            proc = Processor(stream=stream, f=f, name=name or f.__name__, **kwargs)
            logger.debug("found-processor", name=proc.name)

            if proc.id in self.processors:
                raise Misconfigured("Processor name must be unique within an App and Stream")

            self.processors[proc.id] = proc
            return proc

        return decorator

    def _task(self, func, on_leader=False):
        def decorator(f):
            self.tasks.add(f)
            return func

        return decorator(func)

    def task(self, func=None, *, on_leader=False):
        """
        Define an async function to run at worker startup.

        Parameters
        ----------
        on_leader : bool
            Whether to run the function only on one worker: the elected leader.

        Examples
        --------
        >>> @app.task
        >>> async def on_startup():
        ...     print("starting")

        If you want the task to run on only one worker:

        >>> @app.task(on_leader=True)
        >>> async def on_startup():
        ...    print(f"running once")
        """
        def decorator(f):
            async def wrapper():
                if not on_leader or self.is_leader:
                    logger.info("running-task", name=f.__name__)
                    await f()

            logger.debug("found-task", name=f.__name__)
            return self._task(wrapper)

        return decorator(func) if func is not None else decorator

    def timer(self, *, interval: int, on_leader=False):
        """
        Define an async function to be run at periodic intervals.

        Parameters
        -----------
        interval : float (seconds)
            How often the function executes in seconds.
        on_leader : bool
            Whether to run the function only on one worker: the elected leader.

        Examples
        --------
        >>> @app.timer(interval=10)
        >>> async def every_10_seconds():
        ...     print("10 seconds passed")

        If you want the task to run on only one worker:

        >>> app.timer(interval=5, on_leader=True)
        >>> async def every_5_seconds():
        ...     print("5 seconds passed on the leader")
        """
        def decorator(f):
            async def timer_spawner(*args) -> None:
                async with anyio.create_task_group() as tg:
                    while True:
                        await anyio.sleep(interval)
                        if not on_leader or self.is_leader:
                            logger.debug("spawning-timer-task", name=f.__name__)
                            await tg.spawn(f, *args)

            logger.debug("found-timer", name=f.__name__)
            return self._task(timer_spawner)

        return decorator

    def crontab(self, spec: str, *, timezone=None, on_leader=False):
        """
        Define an async function to be run at the fixed times, defined by the Cron
        format (see `<https://crontab.guru/>`_ for examples).

        Parameters
        ----------
        spec : str
            The Cron spec defining fixed times to run the decorated function.
        timezone : tzinfo
            The timezone to be taken into account for the Cron jobs. If not set value
            from :attr:`runnel.settings.Settings.timezone` will be taken.
        on_leader : bool
            Whether to run the function only on one worker: the elected leader.

        Examples
        --------
        >>> app.crontab("45 17 * * *")
        >>> async def every_5_45_pm():
        ...     print("It is 5:45pm UTC")

        If you want the task to run on only one worker:

        >>> app.crontab("45 17 * * *", on_leader=True)
        >>> async def every_5_45_pm():
        ...     print("It is 5:45pm UTC on the leader")

        With a timezone specification:

        >>> @app.crontab("45 17 * * *", timezone=pytz.timezone('GMT'))
        >>> async def every_5_45_pm():
        ...     print("It is 5:45pm in London")
        """
        def decorator(f):
            async def cron_spawner(*args) -> None:
                _tz = self.settings.timezone if timezone is None else timezone
                async with anyio.create_task_group() as tg:
                    while True:
                        await anyio.sleep(seconds_until(spec, _tz))
                        if not on_leader or self.is_leader:
                            logger.debug("spawning-cron-task", name=f.__name__)
                            await tg.spawn(f, *args)

            logger.debug("found-cron", name=f.__name__)
            return self._task(cron_spawner)

        return decorator



from dataclasses import dataclass, replace
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, Iterable, Type, Union

import structlog

from runnel.exceptions import Misconfigured
from runnel.interfaces import Serializer
from runnel.record import Record

if TYPE_CHECKING:
    from runnel.app import App

logger = structlog.get_logger(__name__)


@dataclass(frozen=True)
class Stream:
    """
    A set of partitioned Redis streams, together representing a single logical event
    stream.

    Not intended to be used directly. Use :attr:`runnel.App.stream` instead.
    """
    app: "App"
    name: str
    record: Type[Record]
    partition_by: Union[str, Callable]
    serializer: Serializer  # Not applicable for records with primitive=True.
    hasher: Callable[[Any], int]
    partition_count: int
    partition_size: int

    def __post_init__(self):
        if self.record._primitive and self.serializer:
            raise Misconfigured("Cannot serialize primitive records")

        by = self.partition_by
        if isinstance(by, str) and not by in self.record.__fields__:
            raise Misconfigured("Stream.partition_by must specify a record field")

    def __hash__(self):
        return hash(self.id)

    def clone(self, **kwargs):
        return replace(self, **kwargs)

    @property
    def id(self):
        return f"{self.app.name}.{self.name}"

    @lru_cache
    def partition_key(self, i):
        return f"__strm:{self.id}.{i}"

    @lru_cache
    def all_partition_keys(self):
        return [self.partition_key(i) for i in range(0, self.partition_count)]

    def route(self, key):
        return self.partition_key(self.hash(key))

    def hash(self, key):
        return self.hasher(key) % self.partition_count

    async def send(self, *records: Iterable[Record], stream_ids=None):
        """
        Send records to partitions of the stream, according to their partition keys.

        Parameters
        ----------
        records : Iterable[Record]
            The records to send.
        stream_ids : Optional[Iterable[str]]
            A list of stream_ids corresponding to the records. Must be the same length
            as records. If ``None``, then ``"*"`` will be used for all records. See
            `<https://redis.io/commands/xadd>`_ for more details.
        """
        if not stream_ids:
            stream_ids = ["*" for _ in range(len(records))]
        assert len(stream_ids) == len(records)

        async with await self.app.redis.pipeline() as pipe:
            for record, stream_id in zip(records, stream_ids):
                await pipe.xadd(
                    name=self.route(self._compute_key(record)),
                    entry=self.serialize(record),
                    max_len=self.partition_size,
                    approximate=True,
                    stream_id=stream_id
                )
            await pipe.execute()

    async def read(self, group, consumer, prefetch, timeout, **keys):
        return await self.app.redis.xreadgroup(  # yapf: disable
            group=group,
            consumer_id=consumer,
            count=prefetch,
            block=timeout,
            **keys,
        )

    async def ack(self, *events):
        if len(events) == 1:
            e = events[0]
            await self.app.redis.xack(e.partition.key, e.group, e.xid)
        else:
            keys = {}

            # XACK supports multiple ids in one command, but only per
            # key and consumer group, so we must preprocess the events.
            for e in events:
                if e.partition.key not in keys:
                    keys[e.partition.key] = {}
                if e.group not in keys[e.partition.key]:
                    keys[e.partition.key] = {e.group: []}

                keys[e.partition.key][e.group].append(e.xid)

            async with await self.app.redis.pipeline() as pipe:
                for key, groups in keys.items():
                    for group, xids in groups.items():
                        await pipe.execute_command(*["XACK", key, group, *xids])
                await pipe.execute()
        logger.debug("acked", events=[e.data for e in events])

    def _compute_key(self, record):
        if isinstance(self.partition_by, str):
            return getattr(record, self.partition_by)
        elif isinstance(self.partition_by, Callable):
            return self.partition_by(record)

    def serialize(self, record):
        if self.record._primitive:
            return {k.encode("utf-8"): v for k, v in record.dict().items()}

        value = self.serializer.dumps(record.dict())
        if self.serializer.compressor:
            value = self.serializer.compressor.compress(value)
        return {b"data": value}

    def deserialize(self, value):
        if self.record._primitive:
            value = {k.decode("utf-8"): v for k, v in value.items()}
        else:
            value = value[b"data"]
            if self.serializer and self.serializer.compressor:
                value = self.serializer.compressor.decompress(value)
            value = self.serializer.loads(value)

        return self.record(**value)



from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import anyio
import structlog
from aiostream.aiter_utils import aitercontext

from runnel.exceptions import Misconfigured
from runnel.middleware import Ack, Deserialize, Take

if TYPE_CHECKING:
    from runnel.runner import Runner
    from runnel.types import Partition, Event

logger = structlog.get_logger(__name__)


@dataclass
class Events:
    """
    An async generator which yields Events (or batches of them). This is the object
    passed to user-defined processor functions.

    Examples
    --------
    >>> from runnel import App, Record
    ...
    >>> app = App(name="example")
    ...
    >>> class Order(Record):
    ...     order_id: int
    ...     amount: int
    ...
    >>> orders = app.stream(name="orders", record=Order, partition_by="order_id")
    ...
    >>> @app.processor(orders)
    ... async def printer(events):
    ...     async for order in events.records():
    ...         print(order.amount)
    """
    runner: "Runner"
    partition: "Partition"
    want: str = "events"
    batch_args: Optional[Tuple] = None
    failed: Set["Event"] = field(default_factory=set)
    finalized: anyio.abc.Event = field(default_factory=anyio.create_event)
    agen: AsyncIterator[Union["Event", List["Event"]]] = None

    @property
    def executor(self):
        return self.runner.executor

    @property
    def stream(self):
        return self.executor.stream

    def __call__(self):
        return self

    def take(self, n, within):
        """
        Configure the events generator to yield batches of `n` events (unless `within`
        seconds pass before `n` are ready, in which case yield all pending events).

        Parameters
        ----------
        n : int
            The desired batch size.
        within : int (seconds)
            The duration to wait for the batch size to be reached before yielding.

        Examples
        --------
        >>> @app.processor(orders)
        ... async def printer(events):
        ...     async for orders in events.take(10, within=1).records():
        ...         # Handle orders as an atomic batch!
        ...         assert 1 <= len(orders) <= 10
        ...         print(orders)

        Notes
        -----
        This method is provided for efficiency. It is intended to be used where batch
        processing of events greatly increases your processing speed. For example, if
        you are loading records into a database, you may want to use its bulk import API
        to ingest a batch of records at a time.

        Warning
        -------
        Runnel acks events after every iteration through the event generator loop. When
        using `take`, this means the entire batch will be acked at once. As a result,
        you must process the batch as a single unit atomically. If you iterate over the
        events in a batch one-at-a-time and you fail half-way through, then the entire
        batch will be considered failed (and handled according to your
        :attr:`runnel.constants.ExceptionPolicy`). This will lead to duplicate
        processing if the batch is retried, or dropped events if the batch is ignored.
        """
        if within > self.executor.processor.grace_period:
            raise Misconfigured("Cannot wait longer than grace_period for a batch")

        self.batch_args = (n, within)
        return self

    def records(self):
        """
        Configure the events generator to deserialize events into Record objects.

        Examples
        --------
        >>> from runnel import Record
        ...
        >>> @app.processor(orders)
        ... async def printer(events):
        ...     async for order in events.records():
        ...         assert isinstance(event, Record)
        ...         print(order.amount)

        If this method is omitted, you will iterate over the low-level :class:`runnel.Event`
        objects, which gives you access to the raw data as ``Dict[bytes, bytes]``.

        >>> from runnel import Event
        ...
        >>> @app.processor(orders)
        ... async def printer(events):
        ...     async for event in events:
        ...         assert isinstance(event, Event)
        ...         print(event.data)
        """
        self.want = "records"
        return self

    def __aiter__(self):
        self.agen = self.iter()
        return self.agen

    async def aclose(self):
        assert self.agen, "Cannot close an event generator that is not running"
        await self.agen.aclose()

    async def iter(self):
        async with self.running():
            # Common kwargs to all middleware handlers.
            kwargs = {"events": self}

            async with AsyncExitStack() as stack:
                enter = stack.enter_async_context
                agen = await enter(aitercontext(self.root()))

                # Construct the middleware pipeline.
                if self.batch_args:
                    agen = await enter(aitercontext(Take(*self.batch_args).handler(agen, **kwargs)))

                # User-provided middleware, which must handle and yield either single
                # events or a batch.
                for m in self.executor.processor.middleware:
                    agen = await enter(aitercontext(m.handler(agen, **kwargs)))

                # Acknowledgement handling.
                agen = await enter(aitercontext(Ack().handler(agen, **kwargs)))

                # Automatic deserialisation.
                if self.want == "records":
                    agen = await enter(aitercontext(Deserialize().handler(agen, **kwargs)))

                # It begins.
                async for x in agen:
                    yield x

    async def root(self):
        queue = self.runner.partitions[self.partition]
        timeout = max(0.05, min(1, self.executor.processor.grace_period / 4))

        while self.partition in self.executor.safe_partitions:
            async with anyio.move_on_after(timeout) as scope:
                event = await queue.get()

            if not scope.cancel_called:
                yield event

    @asynccontextmanager
    async def running(self):
        logger.debug("events-started")
        assert not self.finalized.is_set()
        try:
            yield
        finally:
            await self.finalized.set()
            logger.debug("events-finally-ended")



from runnel.app import App
from runnel.constants import ExceptionPolicy
from runnel.events import Events
from runnel.interfaces import Compressor, Middleware, Serializer
from runnel.record import Record
from runnel.stream import Stream
from runnel.types import Event, Partition

__all__ = [
    "App",
    "Record",
    "Stream",
    "Partition",
    "Event",
    "Events",
    "Serializer",
    "Compressor",
    "Middleware",
    "ExceptionPolicy",
]



import importlib
from pathlib import Path

import structlog

logger = structlog.get_logger()


def autodiscover(app, basedir=None):
    """
    Automatically import Python modules matching the pattern in
    :attr:`runnel.settings.Settings.autodiscover`. If the pattern is None or "" then
    autodiscovery will be skipped.

    This is necessary because processors and tasks are registered with apps via
    decorators. When the worker process starts up, these decorators must be executed so
    that the app knows they exist.

    Parameters
    ----------
    app : runnel.App
        The app instance for which to discover processors and tasks.
    basedir : str
        The directory relative to which we will search the filesystem.
    """
    pattern = app.settings.autodiscover  # e.g. "myproj/**/streams.py"

    if pattern:
        for filename in Path(basedir or Path('.')).glob(pattern):
            module = ".".join(filename.parts)[:-3]
            logger.debug("discovered", module=module)
            importlib.import_module(module)



from dataclasses import dataclass

import anyio
import structlog

from runnel.interfaces import Middleware

logger = structlog.get_logger(__name__)


@dataclass
class Take(Middleware):
    n: int
    within: float

    async def handler(self, parent, events, **kwargs):
        send_stream, receive_stream = anyio.create_memory_object_stream(self.n)
        produced, consumed = anyio.create_event(), anyio.create_event()

        async with anyio.create_task_group() as tg:
            await tg.spawn(self._producer, parent, send_stream, produced, consumed)

            while True:
                async with anyio.move_on_after(self.within):
                    await produced.wait()

                while receive_stream._state.buffer:
                    try:
                        yield await self._consume(receive_stream)
                    # https://github.com/python-trio/trio/issues/638#issuecomment-418588615
                    except GeneratorExit:
                        await tg.cancel_scope.cancel()
                        logger.warning("take-middleware-exit")
                        return

                await consumed.set()
                consumed = anyio.create_event()
                produced = anyio.create_event()

    async def _producer(self, parent, send_stream, produced, consumed):
        async for event in parent:
            await send_stream.send(event)

            if len(send_stream._state.buffer) == self.n:
                await produced.set()
                await consumed.wait()

    async def _consume(self, receive_stream):
        batch = []
        for _ in range(self.n):
            try:
                batch.append(await receive_stream.receive_nowait())
            except anyio.WouldBlock:
                logger.debug("took-too-few", have=len(batch), wanted=self.n)
                break
        logger.debug("take-consumed", batch=[e.data for e in batch])
        return batch



import gzip

from runnel.interfaces import Compressor

try:
    import lz4.frame as lz4
except ImportError:
    lz4 = None


class Gzip(Compressor):
    def compress(self, value):
        return gzip.compress(value)

    def decompress(self, value):
        return gzip.decompress(value)


class LZ4(Compressor):
    def compress(self, value):
        return lz4.compress(value)

    def decompress(self, value):
        return lz4.decompress(value)



import signal
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Set

import anyio
import structlog

from runnel import context
from runnel.discovery import autodiscover
from runnel.exceptions import Misconfigured
from runnel.executor import Executor
from runnel.utils import base64uuid

if TYPE_CHECKING:
    from runnel.app import App

logger = structlog.get_logger(__name__)


@dataclass
class Worker:
    """
    Runs processors for a given app. By default, will concurrently spawn an Executor
    task for every Processor known to the app. (The Executor in turn will spawn n tasks
    to actually process events -- one for every partition of the stream it owns.)
    """
    app: "App"
    id: str = field(default_factory=base64uuid)
    executors: Set[Executor] = field(default_factory=set)
    started: bool = False

    def __hash__(self):
        return object.__hash__(self)

    @property
    def leader_key(self):
        return f"__lead:{self.app.name}"

    def start(self, processors="all"):
        """
        The main entrypoint.

        Parameters
        ----------
        processors : Union[str, List[str]]
            If "all", then run all processors known to the app. Otherwise only run those
            named in the processors list.

        Notes
        -----
        For every processor, the first worker to start will create a consumer group for
        in Redis if it does not already exist. It will set the starting ID to "0", which
        means "process the entire stream history". If you want to select a specific
        consumer group starting ID, see :func:`runnel.Processor.reset`.

        Examples
        --------
        >>> from runnel import Worker
        >>> from mymodule import myapp
        ...
        >>> # Run all processors.
        >>> Worker(myapp).start()
        ...
        >>> # Run specific processors.
        >>> Worker(myapp).start(["myproc1", "myproc2"])

        $ # Run named processor starting at specific ID from the shell
        $ runnel processor reset mymodule:myproc --start=12345-0
        $ runnel worker mymodule:myapp --processors=myproc
        """
        anyio.run(self._start, processors, backend="asyncio")

    async def _start(self, processors="all"):
        context.worker_id.set(self.id)
        assert processors == "all" or isinstance(processors, list)

        if self.started:
            raise Misconfigured("Worker already running")
        logger.info("starting-worker", processors=processors)

        # Import all modules likely to contain Runnel objects (e.g. processors, tasks)
        # that must be registered with the app.
        autodiscover(self.app)

        # Load lua scripts.
        for script in (Path(__file__).parent / "lua").glob("*.lua"):
            self.app.scripts[script.stem] = self.app.redis.register_script(script.read_text())

        # Create executors for all chosen processors.
        for proc in self.app.processors.values():
            if processors == "all" or proc.name in processors:
                await proc.prepare()
                self.executors.add(Executor(id=base64uuid(), processor=proc))

        # First leadership election attempt.
        await self._elect_leader()

        self.app.workers.add(self)
        try:
            async with anyio.open_signal_receiver(signal.SIGINT) as signals:
                async with anyio.create_task_group() as tg:
                    # The main executor tasks.
                    for e in self.executors:
                        await tg.spawn(e.start)

                    # Background tasks, e.g. timers, crontabs etc.
                    for t in self.app.tasks:
                        await tg.spawn(t)

                    # Leadership polling.
                    await tg.spawn(self.elect_leader)
                    self.started = True

                    # Allow for graceful shutdown on SIGINT.
                    async for signum in signals:
                        if signum == signal.SIGINT:
                            logger.critical("sigint")
                            await tg.cancel_scope.cancel()
                            return
        finally:
            self.app.workers.remove(self)
            if not self.app.workers:
                self.app.redis.connection_pool.disconnect()
            logger.critical("worker-exit", eids=[e.id for e in self.executors])

    async def elect_leader(self):
        # Poll a key in Redis and elect ourselves the leader if one does not exist. Not
        # very democratic, but all workers must poll frequently, so there can be no
        # disagreement. The lead worker is used to run tasks once across all workers
        # (e.g. timers, crontabs, etc).
        while True:
            logger.debug("leadership-check")
            await self._elect_leader()
            await anyio.sleep(self.app.settings.leadership_poll_interval / 1000)

    async def _elect_leader(self):
        # The leader key is essentially a lock, so we reuse the lock extension Lua
        # script.
        fn = self.app.scripts["lock_extend"]
        px = self.app.settings.leadership_poll_interval * 4
        leader = await fn.execute(keys=[self.leader_key], args=[self.id, px])

        if leader:
            if not context.is_leader.get():
                logger.info("new-leader")
                context.is_leader.set(True)
        else:
            context.is_leader.set(False)



import structlog

from runnel.interfaces import Middleware
from runnel.types import Event

logger = structlog.get_logger(__name__)


class Ack(Middleware):
    async def handler(self, parent, events, **kwargs):
        async for x in parent:
            assert isinstance(x, (Event, list))
            try:
                yield x
            except GeneratorExit:
                logger.warning("marking-failed", x=x)
                # Mark the offending event (or events) against the root generator so
                # that we can clean up according to the exception policy.
                if isinstance(x, Event):
                    events.failed.add(x)
                else:
                    events.failed.update(x)
                raise
            else:
                if isinstance(x, Event):
                    await events.stream.ack(x)
                else:
                    await events.stream.ack(*x)



class RunnelException(Exception):
    pass


class Misconfigured(RunnelException):
    pass


class RebalanceFailed(RunnelException):
    pass


class NoPartitionsAssigned(RunnelException):
    pass


class AcquirePartitionLockFailed(RunnelException):
    pass


class ExtendPartitionLockFailed(RunnelException):
    pass



import json
from dataclasses import dataclass

from runnel.interfaces import Compressor, Serializer

try:
    import orjson
except ImportError:
    orjson = None


@dataclass(frozen=True)
class JSONSerializer(Serializer):
    compressor: Compressor = None

    @staticmethod
    def _default(obj):
        if hasattr(obj, 'isoformat'):
            return obj.isoformat()
        return obj

    def dumps(self, value):
        return json.dumps(
            obj=value,
            default=self._default,
            separators=(',', ':'),
        ).encode("utf-8")

    def loads(self, value):
        return json.loads(value)


@dataclass(frozen=True)
class FastJSONSerializer(Serializer):
    compressor: Compressor = None

    def dumps(self, value):
        return orjson.dumps(value)

    def loads(self, value):
        return orjson.loads(value)


if orjson:
    default: Serializer = FastJSONSerializer()
else:
    default = JSONSerializer()



import math
from contextlib import asynccontextmanager

import anyio
import structlog

logger = structlog.get_logger(__name__)


async def race(*coros):
    """
    Run multiple coroutines and return the result of the first to finish, cancelling the
    rest.
    """
    if not coros:
        raise ValueError("race expects at least one argument")

    send_stream, receive_stream = anyio.create_memory_object_stream()

    async def f(coro):
        await send_stream.send(await coro)

    async with anyio.create_task_group() as tg:
        for coro in coros:
            await tg.spawn(f, coro)

        result = await receive_stream.receive()
        await tg.cancel_scope.cancel()

    return result


@asynccontextmanager
async def create_daemon_task_group():
    async with anyio.create_task_group() as tg:
        try:
            yield tg
        finally:
            await tg.cancel_scope.cancel()


class WatermarkQueue:
    """
    A queue that enables waiting until it has reached above or below a certain size.
    """
    def __init__(self, watermark: int, direction="lt"):
        self.watermark = watermark
        self.direction = direction
        self._event = anyio.create_event()

        s, r = anyio.create_memory_object_stream(max_buffer_size=math.inf)
        self._send_stream, self._receive_stream = s, r

    def qsize(self):
        # XXX: AnyIO doesn't provide an API for getting buffer statistics, so we're
        # using a private attribute.
        return len(self._receive_stream._state.buffer)

    async def put(self, item):
        await self._send_stream.send(item)
        if self.direction == "gt":
            if self.qsize() > self.watermark:
                await self._event.set()
                self._event = anyio.create_event()

    async def get(self):
        item = await self._receive_stream.receive()
        if self.direction == "lt":
            if self.qsize() < self.watermark:
                await self._event.set()
                self._event = anyio.create_event()
        return item

    async def wait_for_watermark(self):
        lt = self.direction == "lt" and self.qsize() >= self.watermark
        gt = self.direction == "gt" and self.qsize() <= self.watermark
        if lt or gt:
            await self._event.wait()

    def __repr__(self):
        return f"<WatermarkQueue qsize={self.qsize()}>"



from pydantic import BaseModel

from runnel.exceptions import Misconfigured


class Record(BaseModel):
    def __init__(self, *args, **kwargs):
        """
        This class is used to specify structured data types to store in streams.

        It is a specialized version of pydantic.BaseModel, see
        https://pydantic-docs.helpmanual.io/ for more information.

        Examples
        --------
        >>> class Order(Record):
        ...     order_id: int
        ...     created_at: datetime
        ...     amount: int
        ...     item_ids: List[int]

        Records can be nested arbitrarily, e.g. `items: List[Item]` where `Item` is
        another record. They will be serialized (according to the serializer/compressor
        settings) and stored as arbitrary bytes as a single value in a Redis stream
        entry.

        Alternatively, you can use the native Redis stream key/value pairs by setting
        primitive=True, e.g.:

        >>> class UserAction(Record, primitive=True):
        ...     user_id: int
        ...     type: str

        This allows you to benefit from optimisations such as delta compression (see
        http://antirez.com/news/128), at the cost of not supporting nested values. Only
        int, float, bool, str, and bytes are allowed.
        """
        super().__init__(*args, **kwargs)

    def __init_subclass__(cls, /, primitive=False, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._primitive = primitive

        if primitive:
            allowed = {int, float, bool, str, bytes}
            if not set(cls.__annotations__.values()) <= allowed:
                raise Misconfigured(f"Primitive record fields must be in {allowed}")



from contextvars import ContextVar

worker_id = ContextVar("worker_id", default=None)
executor_id = ContextVar("executor_id", default=None)
stream = ContextVar("stream", default=None)
partition_id: ContextVar[int] = ContextVar("partition_id", default=None)
is_leader = ContextVar("is_leader", default=False)



import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict

import anyio
import structlog

if TYPE_CHECKING:
    from runnel.stream import Stream
    from runnel.runner import Runner

logger = structlog.get_logger(__name__)


@dataclass
class Partition:
    """
    The internal representation of a partition of the stream.

    Attributes
    ----------
    number : int
        Which numerical partition this is.
    key : str
        The Redis key under which the stream data structure is stored.
    """
    stream: "Stream"
    number: int
    pointer: bytes = b"0-0"
    lock: anyio.abc.Lock = field(default_factory=anyio.create_lock)

    def __hash__(self):
        return object.__hash__(self)

    def __repr__(self):
        return f"<Partition stream={self.stream.name} i={self.number} pointer={self.pointer}>"

    @property
    def key(self):
        return self.stream.partition_key(self.number)

    def reset(self):
        # A new owner for this partition, reset the pointer so that Redis will return
        # values from the pending entries list if any exist.
        self.pointer = b"0-0"


@dataclass
class Event:
    """
    The internal representation of an event in a stream. Will ordinarily by deserialized
    into a Record type before it is acted upon: e.g. via ``async for record in
    events.records():``. This low-level representation is also available if necessary.

    Attributes
    ----------
    partition: Partition
        The partition this event came from.
    xid: bytes
        The Redis stream ID.
    data : Dict[bytes, bytes]
        The keys and values retrieved from the Redis stream.

    Examples
    --------
    >>> @app.processor(order)
    ... async def myproc(events):
    ...     async for event in events:
    ...         print(event.xid)
    """
    runner: "Runner"
    partition: "Partition"
    group: bytes
    xid: bytes
    data: Dict[bytes, bytes]
    recovering: bool

    def __hash__(self):
        return hash((self.partition, self.xid))

    def __repr__(self):
        if os.environ["RUNNEL_TESTING"] == "1":
            return f"<Event data={self.data}>"
        return f"<Event stream={self.partition.stream.name} partition={self.partition.number} xid={self.xid}>"



from datetime import tzinfo
from typing import Union

import pytz
from pydantic import BaseSettings, PyObject

from runnel.constants import ExceptionPolicy


class Settings(BaseSettings):
    """
    This class should not be used directly, but acts as a container for application
    settings, configured via environment variables (e.g. `RUNNEL_LOG_LEVEL=debug`) or as
    kwargs to the App object.

    Many of the settings represent defaults for processors or streams and can be
    overridden at object initialisation time.

    Parameters
    ----------
    redis_url : str
        The URL to pass to aredis.StrictRedis.from_url. Default: ``"127.0.0.1:6379"``.
    log_format : str
        Either ``"json"`` or ``"console"`` to specify the log output format.
    log_level : str
        The minimum log level to display: one of ``"debug"``, ``"info"``, ``"warning"``
        is recommended.
    autodiscover : str
        The pattern for :func:`pathlib.Path.glob` to find modules containing
        Runnel app-decorated functions (e.g. processors, tasks), which the worker must
        import on startup. Will be called relative to current working directory. For
        example, use ``'myproj/**/streams.py'`` to find all modules called 'streams'
        inside the 'myproj' folder. Default ``None``.
    timezone : tzinfo
        The timezone to use by default for ``app.crontab`` schedules. Default: ``pytz.UTC``.
    leadership_poll_interval : int (milliseconds)
        How long to sleep between worker leadership election attempts. Default: ``4000``
    testing : bool
        Whether we are currently running the test suite.
    default_serializer : Serializer
        An object implementing the Serialzer interface which controls how records
        are stored in the Redis streams. Default: ``JSONSerialzer`` unless ``orjson`` is
        installed (use the ``runnel[fast]`` bundle to install automatically), in which
        case ``FastJSONSerializer``.
    default_partition_count : int
        How many partitions to create. Default: ``16``.
    default_partition_size : int
        The max length of each partition. (Implemented approximately via Redis' MAXLEN
        option to XACK.) Represents the size of the buffer in case processors are
        offline or cannot keep up with the event rate. Default: ``50_000``.
    default_hasher: Callable[Any, int]
        A function used to hash a record's partition key to decide which partition to
        send it to. Defaults to md5 unless `xxhash` is installed (use the `runnel[fast]`
        bundle to install automataically)
    default_exception_policy : ExceptionPolicy
        How to handle exceptions raised in the user-provided processor coroutine.

        * ``HALT``: Raise the exception, halting execution of the affected partition.
        * ``QUARANTINE``: Mark the affected partition as poisoned, and continue with others.
        * ``IGNORE``: Suppress the exception and continue processing regardless.

        Default: ``HALT``.
    default_lock_expiry : int (seconds)
        The duration of the lock on stream partitions owned by executors of this
        processor. This controls the worst case lag a partition's events may
        experience since other executors will have to wait acquire the lock in case
        the owner has died. Default: ``120``.
    default_read_timeout : int (milliseconds)
        How long to stay blocked reading from Redis via XREADGROUP. Nothing depends
        on this. Default: ``4000``.
    default_prefetch_count : int
        The maximum number of events to read from Redis per partition owned by an
        executor. (If a single executor owns all 16 partitions in a stream and
        prefetch_count is 10, then 160 events may be read at once.) Purely an
        optimisation. Default: ``8``.
    default_assignment_attempts : int
        How many times to try to complete a rebalance operation (i.e. acquire our
        declared partitions) before giving up. Default: ``32``.
    default_assignment_sleep : float (seconds)
        How long to wait between attempts to complete a rebalance operation.
        Default: ``2``.
    default_grace_period : float (seconds)
        How long to wait for execution to complete gracefully before cancelling it.
        Default: ``8``.
    default_pool_size : int
        How many concurrent connections to make to Redis to read events. Default: ``16``.
    default_join_delay : int (seconds)
        How long to wait after joining before attempting to acquire partitions.
        Intended to mitigate a thundering herd problem of multiple workers joining
        simultaneously and needing to rebalance multiple times. Default: ``2``.
    """
    class Config:
        env_prefix = "RUNNEL_"
        case_insensitive = True

    redis_url: str = "redis://127.0.0.1:6379"
    log_format: str = "console"
    log_level: str = "info"
    autodiscover: str = None
    timezone: tzinfo = pytz.UTC
    leadership_poll_interval: int = 4000
    testing: bool = False

    # Stream defaults
    default_partition_count: int = 16
    default_partition_size = 50_000
    default_serializer: Union[PyObject, str] = "runnel.serialization.default"
    default_hasher: Union[PyObject, str] = "runnel.hashing.default"

    # Processor defaults
    default_exception_policy: ExceptionPolicy = ExceptionPolicy.HALT
    default_lock_expiry: int = 60 * 2
    default_read_timeout: int = 4000
    default_prefetch_count: int = 8
    default_assignment_attempts: int = 32
    default_assignment_sleep: float = 2
    default_grace_period: float = 8
    default_pool_size: int = 16
    default_join_delay: int = 2



from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from functools import lru_cache
from typing import TYPE_CHECKING, Dict, Set

import anyio
import structlog

from runnel.concurrency import WatermarkQueue, race
from runnel.constants import ExceptionPolicy
from runnel.context import partition_id
from runnel.events import Events
from runnel.types import Event, Partition

if TYPE_CHECKING:
    from runnel.executor import Executor

logger = structlog.get_logger(__name__)


@dataclass
class Runner:
    """
    The Runner is responsible for concurrently running a user-provided processor
    function for every partition owned by the Executor. It will pass the function
    an events generator which must be fed events from a queue.
    """
    executor: "Executor"
    partitions: Dict[Partition, WatermarkQueue] = field(default_factory=dict)
    new_joiner: anyio.abc.Event = field(default_factory=anyio.create_event)
    new_leaver: anyio.abc.Event = field(default_factory=anyio.create_event)
    fetch_completed: anyio.abc.Event = field(default_factory=anyio.create_event)
    fetching: Set[Partition] = field(default_factory=set)
    pool: anyio.abc.Semaphore = None

    def __post_init__(self):
        self.pool = anyio.create_semaphore(self.processor.pool_size)

    @property
    def processor(self):
        return self.executor.processor

    @property
    def stream(self):
        return self.executor.stream

    @property
    def prefetch_count(self):
        return self.processor.prefetch_count

    def need_change(self) -> bool:
        return set(self.partitions) < self.executor.safe_partitions

    def with_space(self) -> Set[Partition]:
        return {p for p, q in self.partitions.items() if q.qsize() < self.prefetch_count}

    def can_fetch(self) -> Set[Partition]:
        return set(self.partitions) - self.fetching

    def should_fetch(self) -> Set[Partition]:
        return self.with_space() - self.fetching

    async def wait_queues_ready(self, queues):
        await race(*[q.wait_for_watermark() for q in queues])

    async def wait_until_should_fetch(self):
        coros = [self.new_joiner.wait(), self.fetch_completed.wait()]
        queues = [self.partitions[p] for p in self.can_fetch()]
        if queues:
            coros.append(self.wait_queues_ready(queues))
        await race(*coros)

    @asynccontextmanager
    async def is_running(self, partition: "Partition"):
        async with partition.lock:
            await self.new_joiner.set()
            self.new_joiner = anyio.create_event()
            try:
                yield
            finally:
                logger.debug("consumer-cleanup")
                # Reset the the partition's pointer so that the next task to take
                # ownership starts by reading the pending entries list. This is
                # necessary because any events we have prefetched and are currently
                # queued up will be deleted in the next line.
                partition.reset()
                del self.partitions[partition]
                # Wait until any pending fetches for the partition complete. Otherwise,
                # there's a race condition between the current fetcher and the one that
                # will own the partition next. This fetcher will read, but not ack, and
                # any events will stay in the PEL. Since we read as the same Redis
                # group/consumer, depending on its startup time the new owner might miss
                # the events read by this fetcher.
                while partition in self.fetching:
                    await self.fetch_completed.wait()
                # Notify any waiting tasks that we are about to exit and relinquish our
                # run lock.
                await self.new_leaver.set()
                self.new_leaver = anyio.create_event()
                logger.debug("consumer-exit")

    async def run(self):
        try:
            logger.debug("runner-starting")
            async with anyio.create_task_group() as tg:
                # A background task that fetches events from Redis
                # and puts them in a queue per partition.
                await tg.spawn(self.fetcher)

                while True:
                    # Ensure that all partitions owned by our Executor have running consumer
                    # tasks. They are responsible for running the user-provided processor
                    # function and passing it an events generator.
                    logger.debug("runner-loop", partitions=self.partitions)
                    if not self.need_change():
                        logger.debug("runner-waiting")
                        await race(self.executor.wait_for(self.need_change), self.new_leaver.wait())

                    for p in self.executor.safe_partitions:
                        if p not in self.partitions:
                            self.partitions[p] = WatermarkQueue(self.prefetch_count)
                            logger.debug("spawning-consumer", partition=p)
                            await tg.spawn(self.consumer, p)
        finally:
            logger.debug("runner-exit")

    async def consumer(self, partition: "Partition"):
        # Run the processor function over an events generator for a single partition.
        partition_id.set(partition.number)

        async with self.is_running(partition):
            try:
                logger.debug("starting-consumer")
                events = Events(runner=self, partition=partition)

                async with anyio.create_task_group() as tg:
                    await tg.spawn(self.processor.f, events)
                    await self.executor.wait_for(lambda: partition in self.executor.need_release)

                    # The event generator will close gracefully now on the next
                    # iteration through its read loop when it discovers that its
                    # partition needs release. This graceful exit is important, because
                    # it allows the pipeline to ack all the events it has seen, which
                    # prevents duplicate processing once rebalancing is complete.
                    async with anyio.move_on_after(self.processor.grace_period) as scope:
                        logger.debug("consumer-grace-period")
                        await events.finalized.wait()

                    if scope.cancel_called:
                        # The grace period was exceeded, forcefully cancel the
                        # processor. Duplicate processing of unacked events is likely.
                        logger.critical("consumer-grace-period-exceeded")
                        await tg.cancel_scope.cancel()
            except Exception:
                logger.exception("consumer-failed")
                await events.aclose()
                assert events.finalized and events.failed

                policy = self.processor.exception_policy
                if policy == ExceptionPolicy.HALT:
                    # Propagate the exception, killing this executor and any others who
                    # end up taking ownership of this partition. The nuclear option.
                    raise
                elif policy == ExceptionPolicy.QUARANTINE:
                    # It is assumed that the exception is unrecoverable (e.g. a
                    # deserialization error occurred or user-processor code failed).
                    # Mark the offending partition as poisoned and do not attempt to
                    # read from it again. Manual intervention is required to make
                    # progress on that partition -- at a minimum the bad event must be
                    # deleted.
                    self.executor.poisoned.update({e.partition for e in events.failed})
                elif policy == ExceptionPolicy.IGNORE:
                    # Simply ack the failed events. They will not be processed again.
                    # Can be used with dead-letter-queue middleware to implement an
                    # effective strategy for always making forward progress.
                    await self.executor.stream.ack(*events.failed)
            finally:
                await events.aclose()

    async def fetcher(self):
        # 1) We want to use as few connections as possible (i.e. avoid n connections for
        # n partitions if every consumer fetched its own events).
        # 2) We want to allow slow partitions to fall behind, not impacting others, and
        # not buffering too many events in memory.
        # 3) We want to use the blocking XREADGROUP command (rather than
        # constant polling) for efficiency.
        # These demands explain the fetching scheme implemented below, which spawns
        # concurrent fetch tasks, limited by a semaphore, for partitions with queue
        # space.
        async with anyio.create_task_group() as tg:
            while True:
                partitions = self.should_fetch()
                if partitions:
                    self.fetching.update(partitions)
                    await tg.spawn(self.fetch, partitions)
                else:
                    logger.debug("waiting-until-should-fetch")
                    await self.wait_until_should_fetch()

    async def fetch(self, partitions):
        try:
            async with self.pool:
                await self._fetch(partitions)
        finally:
            self.fetching.difference_update(partitions)
            await self.fetch_completed.set()
            self.fetch_completed = anyio.create_event()

    async def _fetch(self, partitions):
        logger.debug("fetching", partitions=partitions)

        results = await self.executor.stream.read(  # yapf: disable
            group=self.executor.group,
            consumer=self.executor.consumer,
            prefetch=self.executor.processor.prefetch_count,
            timeout=self.executor.processor.read_timeout,
            **{p.key: p.pointer for p in partitions},
        )

        for key, values in results.items():
            partition = self.key_to_partition(key)

            try:
                queue = self.partitions[partition]
            except KeyError:
                # A partition has left since beginning the fetch. (Its consumer could
                # have failed and exited, or it could have been reassigned in a
                # rebalance.)
                logger.debug("fetched-partition-not-owned", partition=partition)
                continue

            for xid, data in values:
                logger.debug("putting", partition=partition.number, xid=xid, data=data)
                await queue.put(
                    Event(
                        runner=self,
                        partition=partition,
                        group=self.executor.group,
                        xid=xid,
                        data=data,
                        recovering=partition.pointer != b">",
                    )
                )

            if partition.pointer == b">":
                continue
            elif values:
                # We are reading the PEL, update the pointer to last seen.
                partition.pointer = xid
            else:
                # We have finished reading from the PEL, recovery over.
                partition.pointer = b">"

    def key_to_partition(self, key):
        return self.executor.partitions[self.key_to_int(key)]

    @staticmethod
    @lru_cache(maxsize=2**12)
    def key_to_int(key):
        return int(key.split(b".")[-1])



import json
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from typing import TYPE_CHECKING, Coroutine, Dict, List

import structlog
from aredis.exceptions import RedisError

from runnel.constants import ExceptionPolicy, Rebalance
from runnel.exceptions import Misconfigured
from runnel.utils import chunks

if TYPE_CHECKING:
    from runnel.stream import Stream
    from runnel.interfaces import Middleware

logger = structlog.get_logger(__name__)


@dataclass(frozen=True)
class Processor:
    """
    A wrapper around a Python function which iterates over a continuous event stream.

    A processor will be run by one or more executors. Each executor will be responsible
    for processing a different subset of partitions of the underlying Redis streams.

    Not intended to be used directly. Use :attr:`runnel.App.processor` instead.
    """
    stream: "Stream"
    f: Coroutine
    name: str
    exception_policy: ExceptionPolicy
    middleware: List["Middleware"]
    lock_expiry: int
    read_timeout: int
    prefetch_count: int
    assignment_attempts: int
    assignment_sleep: float
    grace_period: float
    pool_size: int
    join_delay: int

    def __post_init__(self):
        if self.read_timeout > self.grace_period * 1000:
            raise Misconfigured("read_timeout must be less than grace_period")

    def __hash__(self):
        return hash(self.id)

    async def __call__(self, *args, **kwargs):
        await self.f(*args, **kwargs)

    @property
    def app(self):
        return self.stream.app

    @property
    def id(self):
        return f"{self.stream.id}.{self.name}"

    @property
    def group(self):
        # The consumer group identity in Redis.
        return self.name

    @property
    def consumer(self):
        # The consumer group consumer identity in Redis. (There is only ever
        # a single consumer since our raison d'etre is in-order processing,
        # not fanning-out a single partition to multiple consumers.)
        return "1"

    @property
    def members_key(self):
        return f"__memb:{self.id}"

    @lru_cache
    def heartbeat_key(self, executor_id):
        return f"__beat:{self.id}.{executor_id}"

    @property
    def control_key(self):
        return f"__ctrl:{self.id}"

    def lock_key(self, name):
        assert name == "admin" or isinstance(name, int)
        return f"__lock:{self.id}.{name}"

    @asynccontextmanager
    async def admin_lock(self):
        timeout = self.app.settings.default_lock_expiry
        async with self.app.redis.lock(self.lock_key("admin"), timeout=timeout):
            yield

    async def members(self):
        return await self.app.redis.get(self.members_key)

    async def maybe_join(self, executor_id: str):
        await self.rebalance(Rebalance.JOIN, executor_id)

    async def maybe_leave(self, executor_id: str):
        await self.rebalance(Rebalance.LEAVE, executor_id)

    async def read_control_messages(self, start="$", timeout=4000):
        while True:
            results = await self.app.redis.xread(
                count=1,
                block=timeout,
                **{self.control_key: start},
            )
            for stream, result in results.items():
                assert stream.decode("utf-8") == self.control_key
                for xid, data in result:
                    yield data
                    start = xid

    async def send_control_message(self, **kwargs):
        await self.app.redis.xadd(self.control_key, kwargs, max_len=256)

    async def rebalance(self, reason: Rebalance, executor_id: str):
        async with self.admin_lock():
            members = json.loads(await self.members() or "{}")

            if reason == Rebalance.JOIN and executor_id in members:
                return
            if reason == Rebalance.LEAVE and executor_id not in members:
                return

            new = self._rebalance(reason, members, executor_id)
            logger.info("reassigning-partitions", new=new)
            await self.app.redis.set(self.members_key, json.dumps(new))

        msg = reason.message(executor_ids=json.dumps([executor_id]))
        await self.send_control_message(**msg)

    def _rebalance(self, reason, members: Dict, executor_id):
        executors = {x: members[x]["joined_at"] for x in members}

        if reason == Rebalance.JOIN:
            executors[executor_id] = datetime.utcnow().isoformat()
        elif reason == Rebalance.LEAVE:
            del executors[executor_id]
        else:
            raise ValueError(f"unexpected rebalance reason: {reason}")

        return self._compute_assignments(executors)

    async def reap(self):
        # Check for any executors who exist in the members list, but whose heartbeat has
        # expired -- they are considered dead. They must be removed from the list and
        # their partitions reassigned to other members.
        async with self.admin_lock():
            members = json.loads(await self.members() or "{}")
            if members:
                new = await self._reap(members)
                if new:
                    logger.critical("dead-executor-found", new=new)
                    await self.app.redis.set(self.members_key, json.dumps(new))
                    dead = [x for x in members if x not in new]
                    msg = Rebalance.DEAD.message(executor_ids=dead)
                    await self.app.redis.publish(self.control_key, msg)

    async def _reap(self, members):
        keys = [self.heartbeat_key(x) for x in members]
        heartbeats = await self.app.redis.mget(keys)
        alive = {x: members[x]["joined_at"] for x, found in zip(members, heartbeats) if found}

        if all(k in alive for k in members):
            return
        return self._compute_assignments(alive)

    def _compute_assignments(self, executors: Dict[str, str]):
        # This function takes a mapping from executor_ids to their join times and returns
        # a new members dictionary including the partition assignments. A naive random
        # assignment would result in large changes every time an executor joins. Using
        # consistent hashing would minimise changes, but can suffer from an uneven
        # distribution of partitions. Instead we order executors by their join date and
        # split partitions evenly. This results in more changes than necessary -- O(n)
        # rather than O(m/n) where n=executors m=partitions -- but gives us perfect
        # fairness, which is more important given our small number of partitions.
        return {
            executor_id: {
                "joined_at": executors[executor_id],
                "partitions": chunk,
            }
            for executor_id, chunk in zip(
                sorted(executors, key=lambda e: executors[e]),
                chunks(list(range(0, self.stream.partition_count)), len(executors))
            )
        }

    async def group_setid(self, start):
        keys = self.stream.all_partition_keys()
        cmds = [("XGROUP", "SETID", key, self.group, start) for key in keys]
        async with await self.app.redis.pipeline() as pipe:
            for cmd in cmds:
                await pipe.execute_command(*cmd)
            await pipe.execute()

    async def group_destroy(self):
        keys = self.stream.all_partition_keys()
        cmds = [("XGROUP", "DESTROY", key, self.group) for key in keys]
        async with await self.app.redis.pipeline() as pipe:
            for cmd in cmds:
                await pipe.execute_command(*cmd)
            await pipe.execute()

    async def group_create(self, start="0"):
        keys = self.stream.all_partition_keys()
        cmds = [("XGROUP", "CREATE", key, self.group, start, "MKSTREAM") for key in keys]
        async with await self.app.redis.pipeline() as pipe:
            for cmd in cmds:
                await pipe.execute_command(*cmd)
            await pipe.execute()

    async def group_exists(self):
        keys = self.stream.all_partition_keys()
        try:
            # Just check the last one, since they are either all created or none.
            info = await self.app.redis.xinfo_groups(keys[-1])
            return self.group.encode("utf-8") in {x[b"name"] for x in info}
        except RedisError as e:
            assert str(e) == "no such key"
            return False

    async def prepare(self):
        """
        Run XGROUP CREATE for all partitions of this processor's stream, if they do not
        already exist.
        """
        async with self.admin_lock():
            if not await self.group_exists():
                await self.group_create()

    async def reset(self, start):
        """
        Run XGROUP DESTORY (if the group exists) followed by XGROUP CREATE with the
        given starting id for all partitions of this processor's stream.

        This should be executed once all workers have been shut down. It will purge any
        unacked events from the 'pending entries list' and will start over from scratch.

        Parameters
        ----------
        start : str
            The starting ID argument to XGROUP CREATE, see https://redis.io/commands/xgroup
        """
        async with self.admin_lock():
            if await self.group_exists():
                await self.group_destroy()
            await self.group_create(start)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, AsyncIterator, List, Union

if TYPE_CHECKING:
    from runnel.types import Event


class Compressor:
    """
    Implement this interface to customise event data compression.

    Examples
    --------
    >>> import gzip
    ...
    >>> class Gzip(Compressor):
    ...     def compress(self, value):
    ...         return gzip.compress(value)
    ...
    ...     def decompress(self, value):
    ...         return gzip.decompress(value)

    Then configure your serializer to use it:

    >>> from runnel import App, Record, JSONSerializer
    ...
    >>> app = App(name="example")
    ...
    >>> class Order(Record):
    ...     order_id: int
    ...     amount: int
    ...
    >>> orders = app.stream(
    ...     name="orders",
    ...     record=Order,
    ...     partition_by="order_id",
    ...     serializer=JSONSerializer(compressor=Gzip()),
    ... )
    """
    def compress(self, value: bytes) -> bytes:
        """Return compressed bytes."""
        raise NotImplementedError

    def decompress(self, value: bytes) -> bytes:
        """Return decompressed bytes."""
        raise NotImplementedError


@dataclass(frozen=True)
class Serializer:
    """
    Implement this interface to customise event data serialization.

    Examples
    --------
    >>> import orjson  # A fast JSON library written in Rust.
    ...
    >>> @dataclass(frozen=True)
    >>> class FastJSONSerializer(Serializer):
    >>>     compressor = None
    ...
    ...     def dumps(self, value):
    ...         return orjson.dumps(value)
    ...
    ...     def loads(self, value):
    ...         return orjson.loads(value)

    Then pass it to your stream:

    >>> from runnel import App, Record
    ...
    >>> app = App(name="example")
    ...
    >>> class Order(Record):
    ...     order_id: int
    ...     amount: int
    ...
    >>> orders = app.stream(
    ...     name="orders",
    ...     record=Order,
    ...     partition_by="order_id",
    ...     serializer=FastJSONSerializer(),
    ... )
    """
    compressor: Compressor = None

    def dumps(self, value: Any) -> bytes:
        """Return serialized bytes."""
        raise NotImplementedError

    def loads(self, value: bytes) -> Any:
        """Return deserialized Python object."""
        raise NotImplementedError


class Middleware:
    """
    Middleware are objects with a `handler` method (decribed below). Processors accept
    a list of user-provided middleware.

    The handler methods form a processing pipeline which runs over events before they
    are yielded to the final processor function. Each handler is passed the previous
    handler in the pipeline (called 'parent').

    Examples
    --------
    >>> from runnel.interfaces import Middleware
    ...
    >>> class Printer(Middleware):
    ...     async def handler(self, parent, **kwargs):
    ...         async for x in parent:
    ...             print(x)
    ...             yield x

    It can then be passed to the processor.

    >>> @app.processor(mystream, middleware=[Printer()])
    ... async def proc(events):
    ...     async for record in events.records():
    ...         pass

    Notes
    -----
    Some of Runnel's internal functionality is implemented using middleware. They can be
    found `here <https://github.com/mjwestcott/runnel/tree/master/runnel/middleware>`_.
    """
    async def handler(self, parent, **kwargs) -> AsyncIterator[Union["Event", List["Event"]]]:
        """
        A middleware handler is an async generator. Given a parent generator that yields
        Events or batches of Events, it must yield the same.

        For example:

        >>> async for x in parent:
        ...     assert isinstance(x, (Event, list))
        ...     yield x

        Post-processing logic can be added below, in a finally clause. This allows you
        to respond to errors further up the chain (e.g. in the final processor code):

        >>> async for x in parent:
        ...     try:
        ...         yield x
        ...     finally:
        ...         await logic()

        This is needed because a GeneratorExit exception will be thrown into the yield
        point if an exception is raised in the calling code. (Also note that async
        generator finalization is scheduled by the Python runtime asynchronously, but
        the Runnel framework ensures it has finished before restarting a processor
        function.)

        If you only want to know if an exception was raised further up the chain, you
        can use the following:

        >>> async for x in parent:
        ...     try:
        ...         yield x
        ...     except GeneratorExit:
        ...         await cleanup()
        ...         raise

        (Note: Python requires that GeneratorExit it not ignored, so it must be reraised
        here to avoid a RuntimeError)
        """
        raise NotImplementedError



import json

import anyio
import typer

from runnel.utils import get_object
from runnel.worker import Worker

cli = typer.Typer()
processor = typer.Typer()
cli.add_typer(processor, name="processor")


@cli.command()
def worker(app: str, processors: str = "all"):
    """
    Run a worker for all processors of the given app, or just the processors given in a
    comma-separated string.

    Examples
    --------
    Assuming 'myapp/example.py' contains your Runnel app object:

    .. code-block:: bash

        $ runnel worker myapp.example:myapp

    Or for specific processors:

    .. code-block:: bash

        $ runnel worker myapp.example:myapp --processors=myproc1,myproc2
    """
    obj = get_object(app)
    Worker(obj).start(processors.split(",") if processors != "all" else "all")


@cli.command()
def send(stream: str, value: str):
    """
    Send a given JSON-encoded value to a stream.

    Examples
    --------
    Assuming 'myapp/example.py' contains a stream called 'actions':

    .. code-block:: bash

        $ runnel send myapp.example:actions "{\\"user_id\\": 1, \\"type\\": \\"signup\\"}"
    """
    obj = get_object(stream)
    anyio.run(obj.send, obj.record(**json.loads(value)))


@cli.command()
def sendmany(stream: str, file: typer.FileText):
    """
    Send multiple JSON-encoded values to a stream in a pipelined transaction.

    Examples
    --------
    .. code-block:: bash

        $ echo "{\\"user_id\\": 1, \\"type\\": \\"signup\\"}" >> data.jsonl
        $ echo "{\\"user_id\\": 2, \\"type\\": \\"signup\\"}" >> data.jsonl

    Assuming 'myapp/example.py' contains a stream called 'actions':

    .. code-block:: bash

        $ runnel sendmany myapp.example:actions data.jsonl
    """
    obj = get_object(stream)
    anyio.run(obj.send, *[obj.record(**json.loads(value)) for value in file])


@processor.command()
def reset(name: str, start: str = "0"):
    """
    Reset the processor's starting ID and event backlog.

    Notes
    -----
    This will destroy and recreate the Redis consumer group(s) associated with the
    processor and should be run once all workers have been shut down.

    Examples
    --------
    Assuming 'myapp/example.py' contains a processor called 'printer':

    .. code-block:: bash

        runnel processor reset myapp.example:printer --start="0"
    """
    obj = get_object(name)
    anyio.run(obj.reset, start)



from runnel.middleware.ack import Ack  # noqa
from runnel.middleware.dead import Dead, DeadLetterMiddleware  # noqa
from runnel.middleware.deserialize import Deserialize  # noqa
from runnel.middleware.take import Take  # noqa



import base64
import importlib
import time
import uuid
from contextlib import contextmanager
from datetime import datetime, tzinfo
from typing import Any, Generator

from croniter.croniter import croniter


def base64uuid() -> str:
    return base64.urlsafe_b64encode(uuid.uuid4().bytes).rstrip(b"=").decode("ascii")


@contextmanager
def duration(logger, event: str, **extra: Any) -> Generator:
    logger.info(f"started-{event}", **extra)
    start = time.perf_counter()
    yield
    duration = time.perf_counter() - start
    logger.info(f"ended-{event}", duration=duration, **extra)


def get_object(name: str) -> Any:
    module_name, instance_name = name.split(":", 1)
    module = importlib.import_module(module_name)
    return getattr(module, instance_name)


def chunks(seq, n):
    d, r = divmod(len(seq), n)
    for i in range(n):
        si = (d + 1) * (i if i < r else r) + d * (0 if i < r else i - r)
        yield seq[si:si + (d + 1 if i < r else d)]


def seconds_until(cron_spec, tz: tzinfo = None) -> float:
    """
    Return seconds until next execution given Crontab specification.
    """
    now = time.time()
    now_tz = tz and datetime.now(tz) or now
    return croniter(cron_spec, start_time=now_tz).get_next(float) - now



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['runnel', 'runnel.middleware']

package_data = \
{'': ['*'], 'runnel': ['lua/*']}

install_requires = \
['aiostream>=0.4.1,<0.5.0',
 'anyio>=2.0.0-beta.1,<3.0.0',
 'aredis>=1.1.8,<2.0.0',
 'colorama>=0.4.3,<0.5.0',
 'croniter>=0.3.34,<0.4.0',
 'hiredis>=1.0.1,<2.0.0',
 'pydantic>=1.5.1,<2.0.0',
 'pytz>=2020.1,<2021.0',
 'structlog>=20.1.0,<21.0.0',
 'typer>=0.3.0,<0.4.0']

extras_require = \
{'fast': ['uvloop>=0.14.0,<0.15.0',
          'xxhash>=1.4.4,<2.0.0',
          'orjson>=3.2.1,<4.0.0',
          'lz4>=3.1.0,<4.0.0']}

entry_points = \
{'console_scripts': ['runnel = runnel.cli:cli']}

setup_kwargs = {
    'name': 'runnel',
    'version': '0.1.0a4',
    'description': 'Distributed event processing for Python based on Redis Streams',
    'long_description': None,
    'author': 'Matt Westcott',
    'author_email': 'm.westcott@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
